import sys
import types

import pytest

from mnemosynecore import secrets as secrets_mod
from mnemosynecore import superset as superset_mod


def test_superset_request_success(monkeypatch):
    class Resp:
        def raise_for_status(self):
            return None

        def json(self):
            return {"ok": True}

    captured = {}
    monkeypatch.setattr(
        superset_mod,
        "get_secret",
        lambda conn: {"host": "https://s.example/", "password": "TOKEN"},
    )
    monkeypatch.setattr(
        superset_mod.requests,
        "request",
        lambda method, url, json=None, headers=None: (
            captured.update({"call": (method, url, json, headers)}) or Resp()
        ),
    )

    out = superset_mod.superset_request(
        endpoint="/api/v1/me",
        method="POST",
        payload={"x": 1},
        vault_conn_id="SUP",
    )
    assert out == {"ok": True}
    method, url, payload, headers = captured["call"]
    assert method == "POST"
    assert url == "https://s.example/api/v1/me"
    assert payload == {"x": 1}
    assert headers["Authorization"] == "Bearer TOKEN"


def test_superset_request_raise_for_status(monkeypatch):
    class Resp:
        def raise_for_status(self):
            raise RuntimeError("bad status")

    monkeypatch.setattr(
        superset_mod,
        "get_secret",
        lambda conn: {"host": "https://s.example/", "password": "TOKEN"},
    )
    monkeypatch.setattr(superset_mod.requests, "request", lambda *args, **kwargs: Resp())
    with pytest.raises(RuntimeError, match="bad status"):
        superset_mod.superset_request(endpoint="/x", vault_conn_id="SUP")


def test_get_superset_client_via_superset_api(monkeypatch):
    captured = {}

    def fake_superset_api(**kwargs):
        captured["kwargs"] = kwargs
        return "CLIENT"

    fake_pkg = types.ModuleType("superset_o3_api_lib")
    fake_pkg.superset_o3_api_lib = types.SimpleNamespace(superset_api=fake_superset_api)
    monkeypatch.setitem(sys.modules, "superset_o3_api_lib", fake_pkg)

    monkeypatch.setattr(
        secrets_mod,
        "resolve_secret",
        lambda conn_id, dir_path=None: {
            "host": "https://ss.o3.ru",
            "login": "svc_user",
            "password": "svc_pwd",
            "extra": '{"auth_provider":"ldap"}',
        },
    )

    client = superset_mod.get_superset_client(superset_conn="SUP")
    assert client == "CLIENT"
    assert captured["kwargs"]["username"] == "svc_user"
    assert captured["kwargs"]["auth_provider"] == "ldap"


def test_get_superset_client_fallback_legacy(monkeypatch):
    calls = {}

    class FakeOAuth:
        def __init__(self, **kwargs):
            calls["auth"] = kwargs

    class FakeClient:
        def __init__(self, auth):
            calls["client_auth"] = auth

    fake_pkg = types.ModuleType("superset_o3_api_lib")
    fake_pkg.SupersetAPIClient = FakeClient
    fake_auth = types.ModuleType("superset_o3_api_lib.auth")
    fake_auth.PasswordOAuthFlow = FakeOAuth
    monkeypatch.setitem(sys.modules, "superset_o3_api_lib", fake_pkg)
    monkeypatch.setitem(sys.modules, "superset_o3_api_lib.auth", fake_auth)

    client = superset_mod.get_superset_client(
        superset_conn={
            "host": "https://superset.local",
            "login": "u",
            "password": "p",
            "schema": "client_id",
            "extra": {"client_secret": "sec"},
        }
    )

    assert isinstance(client, FakeClient)
    assert calls["auth"]["superset_host"] == "https://superset.local"
    assert calls["auth"]["client_secret"] == "sec"


def test_superset_dashboard_thumbnail_new_api(monkeypatch):
    class FakeClient:
        def dashboard_thumbnail(self, **kwargs):
            self.kwargs = kwargs
            return b"PNG"

    fake_client = FakeClient()
    monkeypatch.setattr(superset_mod, "get_superset_client", lambda **kwargs: fake_client)

    out = superset_mod.superset_dashboard_thumbnail(
        superset_conn="SUP",
        dashboard_name="Main dashboard",
        dashboard_url="https://ss.o3.ru/superset/dashboard/p/abc/",
        thumb_size=(2048, 1536),
        window_size="2048, 1536",
    )
    assert out == b"PNG"
    assert fake_client.kwargs["dashboard_name"] == "Main dashboard"
    assert fake_client.kwargs["thumb_size"] == "2048,1536"


def test_superset_dashboard_thumbnail_legacy_api(monkeypatch):
    class FakeClient:
        def dashboard_screenshot(self, **kwargs):
            self.kwargs = kwargs
            return b"LEGACY"

    fake_client = FakeClient()
    monkeypatch.setattr(superset_mod, "get_superset_client", lambda **kwargs: fake_client)

    out = superset_mod.superset_dashboard_thumbnail(
        superset_conn="SUP",
        dashboard_id=77,
        thumb_size="800,600",
        window_size=(1600, 1200),
    )
    assert out == b"LEGACY"
    assert fake_client.kwargs["dashboard_id"] == 77
    assert fake_client.kwargs["thumb_size"] == (800, 600)


def test_superset_chart_thumbnail(monkeypatch):
    class FakeClient:
        def chart_thumbnail(self, **kwargs):
            self.kwargs = kwargs
            return b"CHART"

    fake_client = FakeClient()
    monkeypatch.setattr(superset_mod, "get_superset_client", lambda **kwargs: fake_client)

    out = superset_mod.superset_chart_thumbnail(
        superset_conn="SUP",
        chart_id=31729,
        thumb_size="1200,800",
    )
    assert out == b"CHART"
    assert fake_client.kwargs["chart_id"] == 31729


def test_superset_screenshot_dashboard_saves_files(monkeypatch, tmp_path):
    monkeypatch.setattr(superset_mod, "superset_dashboard_thumbnail", lambda **kwargs: b"PNGDATA")

    out = superset_mod.superset_screenshot_dashboard(
        conn_id="SUP",
        dashboards=[
            {"dashboard_name": "Sales main"},
            {"name": "Filters tab", "file_name": "custom_name"},
        ],
        output_dir=str(tmp_path),
    )

    assert len(out) == 2
    assert (tmp_path / "1_Sales_main.png").read_bytes() == b"PNGDATA"
    assert (tmp_path / "2_custom_name.png").read_bytes() == b"PNGDATA"


def test_superset_screenshot_charts_saves_files(monkeypatch, tmp_path):
    monkeypatch.setattr(superset_mod, "superset_chart_thumbnail", lambda **kwargs: b"CHARTDATA")

    out = superset_mod.superset_screenshot_charts(
        conn_id="SUP",
        charts=[{"chart_id": 11}, {"id": 22, "file_name": "finance_chart"}],
        output_dir=str(tmp_path),
    )

    assert len(out) == 2
    assert (tmp_path / "1_chart_11.png").read_bytes() == b"CHARTDATA"
    assert (tmp_path / "2_finance_chart.png").read_bytes() == b"CHARTDATA"


def test_send_superset_dashboard_screenshot(monkeypatch):
    class FakeFiles:
        def upload_file(self, channel_id, files):
            self.channel_id = channel_id
            self.files = files
            return {"file_infos": [{"id": "F1"}]}

    class FakePosts:
        def create_post(self, options):
            self.options = options
            return {"id": "P1"}

    class FakeDriver:
        def __init__(self):
            self.files = FakeFiles()
            self.posts = FakePosts()

    driver = FakeDriver()
    monkeypatch.setattr(superset_mod, "_get_mattermost_driver", lambda bot_id: driver)
    monkeypatch.setattr(superset_mod, "superset_dashboard_thumbnail", lambda **kwargs: b"PNG")

    out = superset_mod.send_superset_dashboard_screenshot(
        channel_id="CH1",
        bot_id="MM_BOT",
        superset_conn="SUP",
        dashboard_name="Revenue",
        text="Daily snapshot",
    )

    assert out["file_id"] == "F1"
    assert out["post_id"] == "P1"
    assert driver.posts.options["file_ids"] == ["F1"]
    assert driver.posts.options["message"] == "Daily snapshot"


def test_send_superset_chart_screenshot(monkeypatch):
    class FakeFiles:
        def upload_file(self, channel_id, files):
            return {"file_infos": [{"id": "F2"}]}

    class FakePosts:
        def create_post(self, options):
            return {"id": "P2"}

    class FakeDriver:
        def __init__(self):
            self.files = FakeFiles()
            self.posts = FakePosts()

    monkeypatch.setattr(superset_mod, "_get_mattermost_driver", lambda bot_id: FakeDriver())
    monkeypatch.setattr(superset_mod, "superset_chart_thumbnail", lambda **kwargs: b"PNG")

    out = superset_mod.send_superset_chart_screenshot(
        channel_id="CH_CHART",
        bot_id="MM_BOT",
        superset_conn="SUP",
        chart_id=123,
    )
    assert out["file_id"] == "F2"
    assert out["post_id"] == "P2"


def test_send_bulk_dashboard_and_chart(monkeypatch):
    monkeypatch.setattr(
        superset_mod,
        "send_superset_dashboard_screenshot",
        lambda **kwargs: {"channel_id": kwargs["channel_id"], "file_id": "FD", "post_id": "PD", "file_name": "a.png"},
    )
    monkeypatch.setattr(
        superset_mod,
        "send_superset_chart_screenshot",
        lambda **kwargs: {"channel_id": kwargs["channel_id"], "file_id": "FC", "post_id": "PC", "file_name": "b.png"},
    )

    dashboards = superset_mod.send_superset_dashboards_to_channels(
        bot_id="MM",
        superset_conn="SUP",
        dashboards=[{"dashboard_name": "Main", "channel_id": "D1"}],
    )
    charts = superset_mod.send_superset_charts_to_channels(
        bot_id="MM",
        superset_conn="SUP",
        charts=[{"chart_id": 1, "channel_id": "C1"}],
    )

    assert dashboards[0]["file_id"] == "FD"
    assert charts[0]["file_id"] == "FC"


def test_superset_chart_thumbnail_test_supports_creds_path(monkeypatch, tmp_path):
    creds_file = tmp_path / "superset_creds.json"
    creds_file.write_text(
        '{"host":"https://ss.local","login":"u","password":"p"}',
        encoding="utf-8",
    )
    captured = {}

    monkeypatch.setattr(
        superset_mod,
        "superset_chart_thumbnail",
        lambda **kwargs: captured.update(kwargs) or b"PNG",
    )

    out = superset_mod.superset_chart_thumbnail_test(
        superset_conn="SUP_CONN",
        chart_id=5,
        creds_path=str(creds_file),
    )
    assert out == b"PNG"
    assert captured["superset_conn"]["host"] == "https://ss.local"


def test_superset_link_builders():
    dashboard_url = superset_mod.superset_dashboard_url(
        superset_conn={"host": "https://ss.local"},
        dashboard_id=123,
    )
    chart_url = superset_mod.superset_chart_url(
        superset_conn={"host": "https://ss.local"},
        chart_id=456,
    )
    assert dashboard_url.endswith("/superset/dashboard/123/")
    assert "slice_id=456" in chart_url

    d_link = superset_mod.superset_dashboard_link(
        superset_conn={"host": "https://ss.local"},
        dashboard_id=123,
        dashboard_name="Main",
    )
    c_link = superset_mod.superset_chart_link(
        superset_conn={"host": "https://ss.local"},
        chart_id=456,
        chart_name="Chart A",
    )
    assert d_link == "[Main](https://ss.local/superset/dashboard/123/)"
    assert c_link == "[Chart A](https://ss.local/superset/explore/?slice_id=456)"


def test_send_superset_dashboard_link(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        __import__("mnemosynecore.mattermost", fromlist=["post_message"]),
        "post_message",
        lambda **kwargs: captured.update(kwargs) or {"post_id": "P1", "root_id": kwargs.get("root_id")},
    )

    out = superset_mod.send_superset_dashboard_link(
        channel_id="CH",
        bot_id="BOT",
        superset_conn={"host": "https://ss.local"},
        dashboard_id=11,
        dashboard_name="Daily KPI",
        text="Header",
        root_id="ROOT1",
    )
    assert out["post_id"] == "P1"
    assert out["root_id"] == "ROOT1"
    assert "[Daily KPI](https://ss.local/superset/dashboard/11/)" in captured["text"]


def test_send_superset_chart_link(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        __import__("mnemosynecore.mattermost", fromlist=["post_message"]),
        "post_message",
        lambda **kwargs: captured.update(kwargs) or {"post_id": "P2", "root_id": kwargs.get("root_id")},
    )

    out = superset_mod.send_superset_chart_link(
        channel_id="CH",
        bot_id="BOT",
        superset_conn={"host": "https://ss.local"},
        chart_id=22,
        chart_name="Orders",
    )
    assert out["post_id"] == "P2"
    assert "[Orders](https://ss.local/superset/explore/?slice_id=22)" in captured["text"]


def test_send_superset_dashboard_links_to_thread(monkeypatch):
    calls = []

    def fake_send_dashboard_link(**kwargs):
        calls.append(kwargs)
        index = len(calls)
        return {
            "post_id": "MAIN" if index == 1 else f"T{index}",
            "root_id": kwargs.get("root_id"),
            "link": "L",
        }

    monkeypatch.setattr(superset_mod, "send_superset_dashboard_link", fake_send_dashboard_link)
    monkeypatch.setattr(
        __import__("mnemosynecore.mattermost", fromlist=["send_thread_messages"]),
        "send_thread_messages",
        lambda **kwargs: {"sent": [{"post_id": "X"}], "failed": []},
    )

    out = superset_mod.send_superset_dashboard_links_to_thread(
        channel_id="CH",
        bot_id="BOT",
        superset_conn="SUP",
        dashboards=[
            {"dashboard_name": "Main", "id": 1},
            {"dashboard_name": "Thread", "id": 2},
        ],
        header_text="Weekly",
        thread_messages=["plain text in thread"],
    )
    assert out["main_post_id"] == "MAIN"
    assert out["root_id"] == "MAIN"
    assert len(out["links_sent"]) == 2
    assert calls[1]["root_id"] == "MAIN"


def test_superset_screenshot_to_file_helpers(monkeypatch, tmp_path):
    monkeypatch.setattr(superset_mod, "superset_dashboard_thumbnail", lambda **kwargs: b"DASH")
    monkeypatch.setattr(superset_mod, "superset_chart_thumbnail", lambda **kwargs: b"CHART")

    dash_path = superset_mod.superset_dashboard_screenshot_to_file(
        superset_conn="SUP",
        dashboard_name="Main",
        output_path=str(tmp_path / "dash.png"),
    )
    chart_path = superset_mod.superset_chart_screenshot_to_file(
        superset_conn="SUP",
        chart_id=12,
        output_path=str(tmp_path / "chart.png"),
    )
    assert (tmp_path / "dash.png").read_bytes() == b"DASH"
    assert (tmp_path / "chart.png").read_bytes() == b"CHART"
    assert dash_path.endswith("dash.png")
    assert chart_path.endswith("chart.png")


def test_send_superset_dashboard_permalink_screenshot(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        superset_mod,
        "send_superset_dashboard_screenshot",
        lambda **kwargs: captured.update(kwargs) or {"post_id": "P1", "file_id": "F1"},
    )

    out = superset_mod.send_superset_dashboard_permalink_screenshot(
        channel_id="CH",
        bot_id="BOT",
        superset_conn="SUP",
        dashboard_url="https://ss.local/superset/dashboard/p/abc/?native_filters_key=xyz",
        dashboard_name="Filtered board",
        text="message",
    )
    assert out["post_id"] == "P1"
    assert captured["dashboard_url"].startswith("https://ss.local/superset/dashboard/p/abc/")
    assert "[Filtered board](https://ss.local/superset/dashboard/p/abc/?native_filters_key=xyz)" in captured["text"]



import logging
import json
from pathlib import Path

import pytest

from mnemosynecore import mattermost


class _FakePosts:
    def __init__(self, should_fail=False):
        self.should_fail = should_fail
        self.calls = []

    def create_post(self, options):
        if self.should_fail:
            raise RuntimeError("boom")
        self.calls.append(options)


class _FakeDriver:
    def __init__(self, cfg, should_fail=False):
        self.cfg = cfg
        self.posts = _FakePosts(should_fail=should_fail)
        self.logged_in = False

    def login(self):
        self.logged_in = True


class _FakeFiles:
    def __init__(self):
        self.calls = []

    def upload_file(self, channel_id, files):
        self.calls.append((channel_id, files))
        return {"file_infos": [{"id": "FILE1"}]}


class _FakeDriverWithFiles:
    def __init__(self):
        self.files = _FakeFiles()
        self.posts = _FakePosts()


def test_get_mattermost_driver_from_json(monkeypatch):
    monkeypatch.setattr(mattermost, "Driver", _FakeDriver)
    raw = '{"host":"h","password":"t","schema":"http","port":8065,"basepath":"/api/v4"}'
    drv = mattermost.get_mattermost_driver(raw)
    assert drv.logged_in is True
    assert drv.cfg["url"] == "h"
    assert drv.cfg["token"] == "t"


def test_get_mattermost_driver_invalid_json():
    with pytest.raises(ValueError, match="Неверный формат JSON"):
        mattermost.get_mattermost_driver("{bad")


def test_get_mattermost_driver_from_conn(monkeypatch):
    monkeypatch.setattr(mattermost, "un_conn", lambda bot_id, conn_type: ("OK", bot_id, conn_type))
    assert mattermost.get_mattermost_driver("MM_CONN") == ("OK", "MM_CONN", "mattermost")


def test_send_message_success(monkeypatch):
    drv = _FakeDriver({})
    monkeypatch.setattr(mattermost, "get_mattermost_driver", lambda bot_id: drv)
    logs = []
    monkeypatch.setattr(logging, "info", lambda msg, cid: logs.append((msg, cid)))

    mattermost.send_message(channel_id="CH", bot_id="BOT", text="  hi  ")

    assert drv.posts.calls == [{"channel_id": "CH", "message": "hi"}]
    assert logs and logs[0][1] == "CH"


def test_post_message_returns_post_id(monkeypatch):
    class _Posts:
        def __init__(self):
            self.calls = []

        def create_post(self, options):
            self.calls.append(options)
            return {"id": "P1"}

    class _Driver:
        def __init__(self):
            self.posts = _Posts()

    drv = _Driver()
    monkeypatch.setattr(mattermost, "get_mattermost_driver", lambda bot_id: drv)

    out = mattermost.post_message(
        channel_id="CH",
        bot_id="BOT",
        text="hello",
        root_id="ROOT1",
        silent=True,
    )
    assert out["post_id"] == "P1"
    assert drv.posts.calls[0]["root_id"] == "ROOT1"


def test_send_message_reraises(monkeypatch):
    drv = _FakeDriver({}, should_fail=True)
    monkeypatch.setattr(mattermost, "get_mattermost_driver", lambda bot_id: drv)
    with pytest.raises(RuntimeError, match="boom"):
        mattermost.send_message(channel_id="CH", bot_id="BOT", text="x")


def test_send_message_test_success(monkeypatch):
    monkeypatch.setattr(mattermost, "Driver", _FakeDriver)
    monkeypatch.setattr(
        __import__("mnemosynecore.vault.client", fromlist=["get_secret_test"]),
        "get_secret_test",
        lambda bot_id, dir_path=None: {"host": "h", "password": "t"},
    )
    mattermost.send_message_test(channel_id="C1", bot_id="B1", text="  test ", silent=True)


def test_send_file_test_supports_creds_path(monkeypatch, tmp_path: Path):
    creds_file = tmp_path / "mm_creds.json"
    creds_file.write_text('{"host":"h","password":"t"}', encoding="utf-8")

    captured = {}
    monkeypatch.setattr(
        mattermost,
        "send_file",
        lambda **kwargs: captured.update(kwargs) or {"file_id": "OK"},
    )

    out = mattermost.send_file_test(
        channel_id="CH",
        bot_id="MM_CONN",
        file_path="/tmp/report.csv",
        creds_path=str(creds_file),
        silent=True,
    )
    assert out["file_id"] == "OK"
    assert json.loads(captured["bot_id"])["host"] == "h"


def test_send_file_bytes(monkeypatch):
    drv = _FakeDriverWithFiles()
    monkeypatch.setattr(mattermost, "get_mattermost_driver", lambda bot_id: drv)

    out = mattermost.send_file_bytes(
        channel_id="CH",
        bot_id="BOT",
        file_name="report.csv",
        file_bytes=b"a,b\n1,2\n",
        text="daily",
        silent=True,
    )

    assert out["file_id"] == "FILE1"
    assert drv.files.calls
    assert drv.posts.calls[0]["file_ids"] == ["FILE1"]


def test_send_file_bytes_to_thread(monkeypatch):
    drv = _FakeDriverWithFiles()
    monkeypatch.setattr(mattermost, "get_mattermost_driver", lambda bot_id: drv)

    out = mattermost.send_file_bytes(
        channel_id="CH",
        bot_id="BOT",
        file_name="report.csv",
        file_bytes=b"1,2",
        root_id="ROOT-X",
        silent=True,
    )
    assert out["root_id"] == "ROOT-X"
    assert drv.posts.calls[0]["root_id"] == "ROOT-X"


def test_send_file_bytes_validates(monkeypatch):
    monkeypatch.setattr(mattermost, "get_mattermost_driver", lambda bot_id: _FakeDriverWithFiles())
    with pytest.raises(ValueError):
        mattermost.send_file_bytes(
            channel_id="CH",
            bot_id="BOT",
            file_name="",
            file_bytes=b"x",
        )
    with pytest.raises(TypeError):
        mattermost.send_file_bytes(
            channel_id="CH",
            bot_id="BOT",
            file_name="a.txt",
            file_bytes="not-bytes",
        )


def test_send_file(monkeypatch, tmp_path: Path):
    path = tmp_path / "f.txt"
    path.write_text("hello", encoding="utf-8")
    captured = {}

    monkeypatch.setattr(
        mattermost,
        "send_file_bytes",
        lambda **kwargs: captured.update(kwargs) or {"file_id": "X", "post_id": "P"},
    )

    out = mattermost.send_file(
        channel_id="CH",
        bot_id="BOT",
        file_path=str(path),
        text="t",
    )
    assert out["file_id"] == "X"
    assert captured["file_name"] == "f.txt"
    assert captured["file_bytes"] == b"hello"


def test_send_file_not_found():
    with pytest.raises(FileNotFoundError):
        mattermost.send_file(channel_id="CH", bot_id="BOT", file_path="/no/such/file.txt")


def test_send_files(monkeypatch):
    calls = []
    monkeypatch.setattr(
        mattermost,
        "send_file",
        lambda **kwargs: calls.append(kwargs["file_path"]) or {"file_id": "X"},
    )

    out = mattermost.send_files(
        channel_id="CH",
        bot_id="BOT",
        file_paths=["a.txt", "b.txt"],
        silent=True,
    )
    assert len(out) == 2
    assert calls == ["a.txt", "b.txt"]


def test_send_dataframe_as_csv(monkeypatch):
    import pandas as pd

    captured = {}
    monkeypatch.setattr(
        mattermost,
        "send_file_bytes",
        lambda **kwargs: captured.update(kwargs) or {"file_id": "X"},
    )
    df = pd.DataFrame([{"a": 1, "b": 2}])

    out = mattermost.send_dataframe_as_csv(
        channel_id="CH",
        bot_id="BOT",
        df=df,
        silent=True,
    )
    assert out["file_id"] == "X"
    assert captured["mime_type"] == "text/csv"
    assert b"a,b" in captured["file_bytes"]


def test_send_dataframe_preview(monkeypatch):
    import pandas as pd

    captured = {}
    monkeypatch.setattr(
        mattermost,
        "send_message",
        lambda **kwargs: captured.update(kwargs),
    )
    df = pd.DataFrame([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    mattermost.send_dataframe_preview(
        channel_id="CH",
        bot_id="BOT",
        df=df,
        title="preview",
        silent=True,
    )
    assert captured["channel_id"] == "CH"
    assert "preview" in captured["text"]
    assert "rows=2" in captured["text"]


def test_send_thread_message(monkeypatch):
    captured = {}
    monkeypatch.setattr(
        mattermost,
        "post_message",
        lambda **kwargs: captured.update(kwargs) or {"post_id": "P1", "root_id": kwargs["root_id"]},
    )

    out = mattermost.send_thread_message(
        channel_id="CH",
        bot_id="BOT",
        root_id="ROOT1",
        text="thread text",
    )
    assert out["root_id"] == "ROOT1"
    assert captured["root_id"] == "ROOT1"


def test_send_thread_messages_continue_on_error(monkeypatch):
    calls = {"i": 0}

    def fake_send_thread_message(**kwargs):
        calls["i"] += 1
        if calls["i"] == 2:
            raise RuntimeError("boom")
        return {"post_id": f"P{calls['i']}", "root_id": kwargs["root_id"]}

    monkeypatch.setattr(mattermost, "send_thread_message", fake_send_thread_message)

    out = mattermost.send_thread_messages(
        channel_id="CH",
        bot_id="BOT",
        root_id="ROOT1",
        messages=["m1", "m2", "m3"],
        continue_on_error=True,
    )
    assert len(out["sent"]) == 2
    assert len(out["failed"]) == 1


def test_send_thread_message_test_uses_test_bot(monkeypatch):
    monkeypatch.setattr(mattermost, "_resolve_test_bot_id", lambda **kwargs: '{"host":"h","password":"t"}')
    captured = {}
    monkeypatch.setattr(
        mattermost,
        "send_thread_message",
        lambda **kwargs: captured.update(kwargs) or {"post_id": "P1", "root_id": kwargs["root_id"]},
    )
    out = mattermost.send_thread_message_test(
        channel_id="CH",
        bot_id="MM_CONN",
        root_id="ROOT2",
        text="ok",
    )
    assert out["root_id"] == "ROOT2"
    assert captured["bot_id"].startswith("{")



import json
import logging
from typing import Any, Dict, Iterable, List, Optional, Union
import pandas as pd
import vertica_python
from mnemosynecore.vault.client import (
    get_secret,
    get_secret_test,
    get_connection_as_json as _get_connection_as_json,
)
from mnemosynecore.vault.univ_conn import un_conn, un_conn_test
from os import listdir, path
HAS_AIRFLOW = False
try:
    from airflow.operators.python import get_current_context
    from airflow import DAG
    from airflow.utils.task_group import TaskGroup
    from airflow.operators.dummy import DummyOperator
    HAS_AIRFLOW = True
except ImportError:
    pass


def read_sql_file(file_path: str):
    if not path.exists(file_path):
        print('Error: no file ' + file_path)
        return None
    with open(file_path, 'r', encoding='utf-8') as f:
        return f.read()


def load_sql_tasks_from_dir(dir_sql: str, vertica_conn_id: str):
    if not HAS_AIRFLOW:
        raise ImportError("Для использования load_sql_tasks_from_dir необходимо установить Airflow")

    from airflow.providers.vertica.operators.vertica import VerticaOperator

    context = get_current_context()
    dag = context['dag']

    tasks = {}
    files_list = listdir(dir_sql)
    for filename in files_list:
        if not filename.endswith(".sql"):
            continue
        path_file = path.join(dir_sql, filename)
        sql = read_sql_file(path_file)
        if sql:
            sql_statements = sql.split(';')
            task_name = 'task_' + filename.replace('.sql', '_vertica')
            tasks[task_name] = VerticaOperator(
                sql=sql_statements,
                task_id=task_name,
                vertica_conn_id=vertica_conn_id,
                dag=dag,
            )
    return tasks


def vertica_dedupe(
        table_name: str,
        unique_keys: Union[str, List[str]],
        conn: Optional[vertica_python.Connection] = None,
        conn_id: Optional[str] = None,
        date_col: Optional[str] = None,
        keep: str = "last",
        commit: bool = True
):

    if isinstance(unique_keys, str):
        unique_keys = [unique_keys]

    close_conn = False
    if conn is None:
        if not conn_id:
            raise ValueError("Нужно указать conn или conn_id")
        conn = vertica_conn(conn_id)
        close_conn = True

    order_clause = f"ORDER BY {date_col} DESC" if date_col and keep == "last" else ""
    unique_cols = ", ".join(unique_keys)

    join_conditions = " AND ".join([f"t.{k} = x.{k}" for k in unique_keys])

    dedupe_sql = f"""
        DELETE FROM {table_name} t
        USING (
            SELECT {unique_cols}, ROW_NUMBER() OVER (PARTITION BY {unique_cols} {order_clause}) AS rn
            FROM {table_name}
        ) x
        WHERE {join_conditions}
          AND x.rn > 1
    """

    vertica_sql(conn=conn, sql=dedupe_sql, commit=commit)

    if close_conn:
        conn.close()

    print(f"Удаление дубликатов из {table_name} завершено по ключам: {unique_keys}")


def get_connection_as_json(conn_name):
    return _get_connection_as_json(conn_name)


def vertica_upsert(
        df: pd.DataFrame,
        table_name: str,
        unique_keys: Union[str, List[str]],
        conn=None,
        conn_id: Optional[str] = None,
        date_col: Optional[str] = None,
        days_back: Optional[int] = None,
        commit: bool = True
):
    if df.empty:
        print("Нет данных для обновления.")
        return

    if isinstance(unique_keys, str):
        unique_keys = [unique_keys]

    close_conn = False
    if conn is None:
        if not conn_id:
            raise ValueError("conn или conn_id обязателен")
        conn = vertica_conn(conn_id)
        close_conn = True

    temp_table = f"{table_name}_tmp"
    create_temp_sql = f"""
        DROP TABLE IF EXISTS {temp_table};
        CREATE LOCAL TEMP TABLE {temp_table} AS
        SELECT * FROM {table_name} WHERE 1=0;
    """
    vertica_sql(conn=conn, sql=create_temp_sql, commit=False)

    rows = [tuple(x) for x in df.to_numpy()]
    cols = ", ".join(df.columns)
    placeholders = ", ".join(["%s"] * len(df.columns))

    insert_sql = f"INSERT INTO {temp_table} ({cols}) VALUES ({placeholders})"
    vertica_sql(conn=conn, sql=insert_sql, params=rows, many=True, commit=False)

    if date_col and days_back:
        delete_sql = f"""
            DELETE FROM {table_name}
            WHERE {date_col} >= CURRENT_DATE - INTERVAL '{days_back} day';
        """
        vertica_sql(conn=conn, sql=delete_sql, commit=False)

    merge_conditions = " AND ".join([f"t.{k} = s.{k}" for k in unique_keys])
    update_assignments = ", ".join([f"{c} = s.{c}" for c in df.columns if c not in unique_keys])

    merge_sql = f"""
        MERGE INTO {table_name} t
        USING {temp_table} s
        ON {merge_conditions}
        WHEN MATCHED THEN UPDATE SET {update_assignments}
        WHEN NOT MATCHED THEN INSERT ({cols}) VALUES ({cols});
    """
    vertica_sql(conn=conn, sql=merge_sql, commit=False)

    if commit:
        conn.commit()

    if close_conn:
        conn.close()

    print(f"Обновление таблицы {table_name} завершено. {len(df)} строк обработано.")


def _build_vertica_connect_kwargs(cfg: Dict[str, Any]) -> Dict[str, Any]:
    extra = cfg.get("extra") or {}
    if isinstance(extra, str):
        try:
            extra = json.loads(extra)
        except Exception:
            extra = {}

    login = cfg.get("user") or cfg.get("login")
    kwargs = {
        "host": cfg["host"],
        "port": int(cfg.get("port", 5433)),
        "user": login,
        "password": cfg.get("password"),
        "database": cfg.get("schema") or cfg.get("database"),
    }
    kwargs.update(extra if isinstance(extra, dict) else {})
    return kwargs


def vertica_conn(conn_id: str):
    cfg = get_secret(conn_id)
    kwargs = _build_vertica_connect_kwargs(cfg)
    return vertica_python.connect(**kwargs)


def vertica_conn_test(
    conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    cfg = get_secret_test(conn_id, dir_path=dir_path, creds_path=creds_path)
    kwargs = _build_vertica_connect_kwargs(cfg)
    return vertica_python.connect(**kwargs)


def get_vertica_engine(conn_id: str):
    return un_conn(conn_id, "vertica")


def get_vertica_engine_test(
    conn_id: str,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    return un_conn_test(
        conn_id=conn_id,
        conn_type="vertica",
        dir_path=dir_path,
        creds_path=creds_path,
    )


def vertica_sql(
    *,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    sql: str,
    params=None,
    many: bool = False,
    commit: bool = True
):
    if not conn and not conn_id:
        raise ValueError("Нужно указать conn или conn_id")

    close_conn = False
    if not conn:
        conn = vertica_conn(conn_id)
        close_conn = True

    try:
        with conn.cursor() as cur:
            if many:
                cur.executemany(sql, params)
            else:
                cur.execute(sql, params)

        if commit:
            conn.commit()

    except Exception:
        conn.rollback()
        logging.exception("Ошибка выполнения SQL в Vertica")
        raise

    finally:
        if close_conn:
            conn.close()


def vertica_select(*, conn_id: str = None, conn=None, sql: str, params=None):
    if not conn and not conn_id:
        raise ValueError("Нужно указать conn или conn_id")

    close_conn = False
    if not conn:
        conn = vertica_conn(conn_id)
        close_conn = True

    try:
        with conn.cursor() as cur:
            cur.execute(sql, params)
            columns = [desc[0] for desc in cur.description]
            rows = cur.fetchall()
        import pandas as pd
        return pd.DataFrame(rows, columns=columns)

    finally:
        if close_conn:
            conn.close()


def split_sql_statements(sql_text: str) -> List[str]:
    if not sql_text:
        return []
    return [q.strip() for q in sql_text.split(";") if q.strip()]


def vertica_sql_file(
    *,
    file_path: str,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    params=None,
    commit: bool = True,
) -> int:
    sql_text = read_sql_file(file_path)
    if sql_text is None:
        raise FileNotFoundError(file_path)

    statements = split_sql_statements(sql_text)
    if not statements:
        return 0

    close_conn = False
    if conn is None:
        if not conn_id:
            raise ValueError("Нужно указать conn или conn_id")
        conn = vertica_conn(conn_id)
        close_conn = True

    try:
        for statement in statements:
            vertica_sql(conn=conn, sql=statement, params=params, commit=False)
        if commit:
            conn.commit()
    finally:
        if close_conn:
            conn.close()
    return len(statements)


def vertica_sql_dir(
    *,
    dir_sql: str,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    recursive: bool = False,
    commit: bool = True,
) -> Dict[str, int]:
    if not path.isdir(dir_sql):
        raise NotADirectoryError(dir_sql)

    sql_files: List[str] = []
    if recursive:
        for root, _, files in __import__("os").walk(dir_sql):
            for filename in files:
                if filename.endswith(".sql"):
                    sql_files.append(path.join(root, filename))
    else:
        for filename in listdir(dir_sql):
            if filename.endswith(".sql"):
                sql_files.append(path.join(dir_sql, filename))

    sql_files.sort()

    close_conn = False
    if conn is None:
        if not conn_id:
            raise ValueError("Нужно указать conn или conn_id")
        conn = vertica_conn(conn_id)
        close_conn = True

    result: Dict[str, int] = {}
    try:
        for sql_file in sql_files:
            executed = vertica_sql_file(
                file_path=sql_file,
                conn=conn,
                commit=False,
            )
            result[sql_file] = executed
        if commit:
            conn.commit()
    finally:
        if close_conn:
            conn.close()

    return result


def vertica_select_scalar(
    *,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    sql: str,
    params=None,
    default=None,
):
    if not conn and not conn_id:
        raise ValueError("Нужно указать conn или conn_id")

    close_conn = False
    if not conn:
        conn = vertica_conn(conn_id)
        close_conn = True

    try:
        with conn.cursor() as cur:
            cur.execute(sql, params)
            row = cur.fetchone()
            if not row:
                return default
            return row[0]
    finally:
        if close_conn:
            conn.close()


def vertica_insert_dataframe(
    *,
    df: pd.DataFrame,
    table_name: str,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    commit: bool = True,
    chunk_size: int = 5000,
) -> int:
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df должен быть pandas.DataFrame")
    if df.empty:
        return 0
    if chunk_size < 1:
        raise ValueError("chunk_size должен быть >= 1")

    close_conn = False
    if conn is None:
        if not conn_id:
            raise ValueError("Нужно указать conn или conn_id")
        conn = vertica_conn(conn_id)
        close_conn = True

    cols = ", ".join(df.columns)
    placeholders = ", ".join(["%s"] * len(df.columns))
    insert_sql = f"INSERT INTO {table_name} ({cols}) VALUES ({placeholders})"
    rows = [tuple(x) for x in df.to_numpy()]

    try:
        for i in range(0, len(rows), chunk_size):
            chunk = rows[i:i + chunk_size]
            vertica_sql(conn=conn, sql=insert_sql, params=chunk, many=True, commit=False)
        if commit:
            conn.commit()
    finally:
        if close_conn:
            conn.close()

    return len(rows)


def vertica_table_exists(
    *,
    table_name: str,
    schema_name: Optional[str] = None,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
) -> bool:
    if not conn and not conn_id:
        raise ValueError("Нужно указать conn или conn_id")

    close_conn = False
    if not conn:
        conn = vertica_conn(conn_id)
        close_conn = True

    try:
        with conn.cursor() as cur:
            if schema_name:
                cur.execute(
                    """
                    SELECT 1
                    FROM v_catalog.tables
                    WHERE table_schema = %s AND table_name = %s
                    LIMIT 1
                    """,
                    (schema_name, table_name),
                )
            else:
                cur.execute(
                    """
                    SELECT 1
                    FROM v_catalog.tables
                    WHERE table_name = %s
                    LIMIT 1
                    """,
                    (table_name,),
                )
            return cur.fetchone() is not None
    finally:
        if close_conn:
            conn.close()


def _resolve_vertica_test_conn(
    *,
    conn=None,
    conn_id: Optional[str] = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    if conn is not None:
        return conn, False
    if not conn_id:
        raise ValueError("Нужно указать conn или conn_id")
    return (
        vertica_conn_test(
            conn_id=conn_id,
            dir_path=dir_path,
            creds_path=creds_path,
        ),
        True,
    )


def vertica_sql_test(
    *,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    sql: str,
    params=None,
    many: bool = False,
    commit: bool = True,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_sql(
            conn=conn,
            sql=sql,
            params=params,
            many=many,
            commit=commit,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_select_test(
    *,
    conn_id: str = None,
    conn=None,
    sql: str,
    params=None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_select(conn=conn, sql=sql, params=params)
    finally:
        if close_conn:
            conn.close()


def vertica_select_scalar_test(
    *,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    sql: str,
    params=None,
    default=None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_select_scalar(
            conn=conn,
            sql=sql,
            params=params,
            default=default,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_sql_file_test(
    *,
    file_path: str,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    params=None,
    commit: bool = True,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> int:
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_sql_file(
            file_path=file_path,
            conn=conn,
            params=params,
            commit=commit,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_sql_dir_test(
    *,
    dir_sql: str,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    recursive: bool = False,
    commit: bool = True,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> Dict[str, int]:
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_sql_dir(
            dir_sql=dir_sql,
            conn=conn,
            recursive=recursive,
            commit=commit,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_insert_dataframe_test(
    *,
    df: pd.DataFrame,
    table_name: str,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    commit: bool = True,
    chunk_size: int = 5000,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> int:
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_insert_dataframe(
            df=df,
            table_name=table_name,
            conn=conn,
            commit=commit,
            chunk_size=chunk_size,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_table_exists_test(
    *,
    table_name: str,
    schema_name: Optional[str] = None,
    conn_id: str = None,
    conn: vertica_python.Connection = None,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
) -> bool:
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_table_exists(
            table_name=table_name,
            schema_name=schema_name,
            conn=conn,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_dedupe_test(
    table_name: str,
    unique_keys: Union[str, List[str]],
    conn: Optional[vertica_python.Connection] = None,
    conn_id: Optional[str] = None,
    date_col: Optional[str] = None,
    keep: str = "last",
    commit: bool = True,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_dedupe(
            table_name=table_name,
            unique_keys=unique_keys,
            conn=conn,
            date_col=date_col,
            keep=keep,
            commit=commit,
        )
    finally:
        if close_conn:
            conn.close()


def vertica_upsert_test(
    df: pd.DataFrame,
    table_name: str,
    unique_keys: Union[str, List[str]],
    conn=None,
    conn_id: Optional[str] = None,
    date_col: Optional[str] = None,
    days_back: Optional[int] = None,
    commit: bool = True,
    dir_path: Optional[str] = None,
    creds_path: Optional[str] = None,
):
    conn, close_conn = _resolve_vertica_test_conn(
        conn=conn,
        conn_id=conn_id,
        dir_path=dir_path,
        creds_path=creds_path,
    )
    try:
        return vertica_upsert(
            df=df,
            table_name=table_name,
            unique_keys=unique_keys,
            conn=conn,
            date_col=date_col,
            days_back=days_back,
            commit=commit,
        )
    finally:
        if close_conn:
            conn.close()
